






/*
 *
 *          Copyright (C) 1995, M. A. Sridhar
 *  
 *
 *     This software is Copyright M. A. Sridhar, 1995. You are free
 *     to copy, modify or distribute this software  as you see fit,
 *     and to use  it  for  any  purpose, provided   this copyright
 *     notice and the following   disclaimer are included  with all
 *     copies.
 *
 *                        DISCLAIMER
 *
 *     The author makes no warranties, either expressed or implied,
 *     with respect  to  this  software, its  quality, performance,
 *     merchantability, or fitness for any particular purpose. This
 *     software is distributed  AS IS.  The  user of this  software
 *     assumes all risks  as to its quality  and performance. In no
 *     event shall the author be liable for any direct, indirect or
 *     consequential damages, even if the  author has been  advised
 *     as to the possibility of such damages.
 *
 */

#if defined(__GNUC__)
#pragma implementation
#endif

#define MNEMONIC_CHAR '&'  // For alt- functions

#   include <iostream.h>
#   include "ui/dsplsurf.h"
#   include "ui/dsinmem.h"
#   include "ui/shadorec.h"
#   include "ui/shadotrg.h"
#   include "ui/graphic.h"

#include <X11/cursorfont.h>

#include "ui/menu.h"
#include "ui/cntroler.h"
#include "ui/composit.h"
#include "ui/font.h"

#include "base/binding.h"
#include "base/treewalk.h"

typedef UI_EventBinding1 <UI_MenuBar> MenuBinding1 ;
template class UI_EventBinding1 <UI_MenuBar> ;
typedef UI_EventBinding1 <UI_MenuItem> MenuBinding2 ;
template class UI_EventBinding1 <UI_MenuItem> ;


UI_MenuItem::UI_MenuItem (UI_Menu* container, const char*
                          label, UI_ViewID id)
: UI_SimpleVObject ((UI_CompositeVObject*) container->Parent(),
    // --------------^^^^^^^^^^^^^^^^^^^^
    // We can safely cast down, because the constructor of UI_Menu insists
    // on a Composite for its parent.
                    UI_Rectangle (), id,  -1),                    
    _container (container)
{
    _SetTitle (label);
    _ownModel   = FALSE;
    _checkState = NotCheckable;
    _focusItem = NULL ;
    _childCount = 0 ;

    // We don't want menu items to inherit the parent's font. So:
    _font = & (container->Font());
}


UI_WindowClass UI_MenuItem::WindowClass () const
{
    // This method is not used for widget creation, only for run-time type
    // identification
    return X_YACL_MENUITEM_CLASS;
}

bool UI_MenuItem::MakeVisualElement ()
{
    return _window != 0 ? TRUE : FALSE;
}


void UI_MenuItem::_DoFocusEmphasis ()
{
  // Do nothing
}


bool UI_MenuItem::HandleEvent (UI_Event* e)
{
    switch(e->Type()) {
    case Event_Paint:
    {
        _DrawItemWindow() ;
        return TRUE ;
    }
    default:
        return ProcessEvent(e) ;
    }
    return ProcessEvent(e) ;
}

void UI_MenuItem::_NotifyGetFocus()
{
    UI_Event evt(Event_GetFocus, this, this) ;
    _Controller -> DispatchEvent(&evt) ;

}

void UI_MenuItem::_NotifyLoseFocus()
{
    UI_Event evt(Event_LoseFocus,this,this) ;
    _Controller -> DispatchEvent(&evt) ;

}

void UI_MenuItem::_ButtonPress(int x,int y)
{
    UI_MenuItem* curr = _CurrentFocusItem(UI_Point(x,y));

    if (!curr) 
        return ;
    
    if ( ! (curr -> _enabled) ) 
        return ;

    if (_focusItem) {
        if (curr -> ViewID() == _focusItem -> ViewID())
            return ;
        _ItemEraseShadowRect(_focusItem) ;
        _focusItem -> _ItemLoseFocus() ;
    } 
    _ItemGetFocus(curr) ;
    _focusItem = curr ;
}

void UI_MenuItem::_ButtonRelease(int x,int y)
{
    UI_MenuItem* curr = _CurrentFocusItem(UI_Point(x,y));

    if (!curr) {
        _container -> _Processed() ;
        return ;
    }
    
    if (_container -> ChildCount(curr -> ViewID()) > 0)
        return ;

    _container -> _Processed() ;

    if (curr -> _enabled) {
        UI_Event evt(Event_Select,curr,curr) ;
        _Controller -> DispatchEvent(&evt) ;
    } 

}

void UI_MenuItem::_MouseMove(int x,int y)
{
    UI_MenuItem* curr = _CurrentFocusItem(UI_Point(x,y));

    if (!curr) {
        if (_focusItem) {
            _ItemEraseShadowRect(_focusItem) ;
            _focusItem -> _ItemLoseFocus() ;
            _focusItem = NULL ;
        }
        return ;
    }
 
    if (!curr -> _enabled) {
        if (_focusItem) {
            _ItemEraseShadowRect(_focusItem) ;
            _focusItem -> _ItemLoseFocus() ;
            _focusItem = NULL ;
        }
        return ;
    }

    if (_focusItem) {
        if (_focusItem -> ViewID() == curr -> ViewID())
            return ;

        _ItemEraseShadowRect(_focusItem) ;
        _focusItem -> _ItemLoseFocus() ;
    }

    _ItemGetFocus(curr) ;
    _focusItem = curr ;
}

UI_MenuItem* UI_MenuItem::_CurrentFocusItem(UI_Point pt)
{

    UI_MenuItem* child ;
    int maxVal = 2;
    int y = pt.YCoord() ;
    int i;
    int incrHt = 0;

    for (i = 0; i < _childCount; i++) {
        child = _container->Child(ViewID(),i) ;

        if (child -> ViewID() >= 32000) 
            incrHt = UI_Menu::SEPARATOR_HEIGHT
                + 2 * UI_Menu::SEPARATOR_VERT_SPACE 
                - UI_Menu::ITEM_VERT_SPACE ;
        else
            incrHt = UI_Menu::_charHeight + UI_Menu::ITEM_VERT_SPACE  ;

        maxVal += incrHt ;
        if (y < maxVal)
	    break;
    }

    if (y <= 2 || i >= _childCount)
        return NULL ;

    UI_MenuItem* retItem ;

    retItem = _container -> Child(ViewID(),i) ;
    if (retItem -> ViewID() >= 32000)
        return NULL ;
        
    return retItem ;
}

bool UI_MenuItem::_DrawItemWindow()
{
    UI_Rectangle rect(0,0,_shape.Width(),_shape.Height());

    UI_DwgSurfaceInMemory memDS(rect.Width(),
                                rect.Height());
    memDS.ColorRectangle(rect,_bgColor);

    int child_count = _childCount ;

    UI_ShadowRectangle temp(0,0, rect.Width(),rect.Height(),
             UI_ShadowRectangle::Raised,2,FALSE,_bgColor);

    temp.DrawOn(memDS);

    UI_MenuItem* item ;

    for (int i = 0; i < child_count; i++)
    {
        item = _container -> Child(ViewID(),i) ;
        UI_Rectangle itemShape = item -> _itemShape ;

        if (item -> ViewID() >= 32000) {
            UI_ShadowRectangle temp(UI_Menu::SEPARATOR_LSPACE,
                itemShape.Top(),
                itemShape.Width(), 
                itemShape.Height(),
                UI_ShadowRectangle::Recessed,1,FALSE,_bgColor);
            temp.DrawOn(memDS);
        }

        else {
            CL_String label = item -> Title() ;
            short mnemonic_pos = label.CharIndex (MNEMONIC_CHAR);
            if (mnemonic_pos >= 0) {
                label(mnemonic_pos,1) = "";
                if (label[mnemonic_pos] == MNEMONIC_CHAR)
                    mnemonic_pos = -1 ;
            }

            UI_Color memDSColor ;
            if (!(item -> _enabled)) {
                memDSColor = memDS.Pen().Color() ;
                memDS.Pen().Color(DisabledForeground()) ;
            }

            memDS.WriteString (label, itemShape) ;

            if (item -> _checkState != NotCheckable)
                _DrawToggleState(memDS,item -> _checkState,itemShape) ;

            if (mnemonic_pos >= 0 && mnemonic_pos < label.Size())
                item -> _DrawUnderLine(memDS,label,
                    mnemonic_pos,itemShape) ;

            if (!(item -> _enabled))
                memDS.Pen().Color(memDSColor) ;
        }

        if (_container->ChildCount(item -> ViewID()) > 0)
        {
             UI_Point pt(rect.Width()  - UI_Menu::TRIANGLE_LFROMR - 
                         UI_Menu::TRIANGLE_WIDTH, 
                         itemShape.Top() + UI_Menu::TRIANGLE_TOP) ;
             UI_ShadowedTriangle tempTri(
                   UI_Rectangle(pt,UI_Menu::TRIANGLE_WIDTH,
                   UI_Menu::TRIANGLE_HEIGHT),
                   UI_ShadowedTriangle::East, UIRelief_Raised);
             tempTri.DrawOn(memDS);
        }
    }

    UI_DisplaySurface& sfc = CreateDisplaySurface();
    memDS.CopyTo(sfc);
    DestroyDisplaySurface();

    return TRUE ;
}

bool UI_MenuItem::_DrawUnderLine (UI_DrawingSurface& sfc,
          CL_String label,short mnemonic_pos,UI_Rectangle rect)
{
    int x1,y1,x2,y2 ;
    UI_Font* font = OurFont() ;
    int charWidth = font->TextWidth(label(mnemonic_pos,1)) ;
    int textWidth = font->TextWidth(label(0,mnemonic_pos)) ;
    x1 = rect.Left() + textWidth ;
    y1 = rect.Top() + UI_Menu::_charHeight + 1;
    x2 = x1 + charWidth;
    y2 = y1 ;
    sfc.DrawLine(UI_Point(x1,y1),UI_Point(x2,y2));
    return TRUE ;
}

bool UI_MenuItem::_DrawToggleState(UI_DrawingSurface& sfc,
                  CheckState state,UI_Rectangle rect)
{
    int xCoord = rect.Left() - (3 * UI_Menu::ITEM_LEFT_SPACE / 2) ;
    int yCoord = rect.Top() + 2 ;

    if (state == NotCheckable)
        return FALSE ;

    if (state == Checked) {
        UI_ShadowRectangle toggleRect (xCoord,yCoord,12,12,
                UI_ShadowRectangle::Recessed,2,FALSE) ;
        toggleRect.DrawOn(sfc) ;
        UI_Rectangle redRect(xCoord+3,yCoord+3,6,6) ;
        sfc.ColorRectangle(redRect,UI_Color(0.804,0.361,0.361)) ;
    }
    else {
        UI_ShadowRectangle toggleRect (xCoord,yCoord,12,12,
                UI_ShadowRectangle::Raised,2,FALSE) ;
        toggleRect.DrawOn(sfc) ;
    }

    return TRUE ;
}

bool UI_MenuItem::_ItemGetFocus(UI_MenuItem* focus)
{
     if (!(focus && !(focus -> IsVisible()) && focus -> ViewID() < 32000))
         return FALSE ;

     UI_DisplaySurface& sfc = CreateDisplaySurface() ;
     UI_DwgSurfaceInMemory memDS
         (Shape().Width() - (UI_Menu::SHADOW_WINDOW_SPACE * 2),
          focus -> _itemShape.Height() + UI_Menu::ITEM_VERT_SPACE,
          &sfc) ;
     memDS.ColorRectangle(memDS.DrawingArea(),_bgColor);

     UI_ShadowRectangle temp(0, 0,
         _shape.Width() - (UI_Menu::SHADOW_WINDOW_SPACE * 2),
         (focus -> _itemShape).Height() + UI_Menu::ITEM_VERT_SPACE,
         UI_ShadowRectangle::Raised,2,FALSE,_bgColor);

     temp.DrawOn(memDS);

     CL_String label = focus -> Title() ;
     short mnemonic_pos = label.CharIndex (MNEMONIC_CHAR);
     if (mnemonic_pos >= 0) {
         label(mnemonic_pos,1) = "";
        if (label[mnemonic_pos] == MNEMONIC_CHAR)
            mnemonic_pos = -1 ;
     }

     int item_left_space = UI_Menu::ITEM_LEFT_SPACE ;

     if (focus -> _checkState != NotCheckable)
         item_left_space *= 2 ;
     
     UI_Rectangle rect (item_left_space - UI_Menu::SHADOW_WINDOW_SPACE, 
                        (UI_Menu::ITEM_VERT_SPACE / 2) ,
                        focus -> _itemShape.Width(),
                        focus -> _itemShape.Height());
     memDS.WriteString (label, rect) ;

     _DrawToggleState (memDS,focus->_checkState,rect) ;

     if (mnemonic_pos >= 0 && mnemonic_pos < label.Size())
         focus -> _DrawUnderLine(memDS,label,mnemonic_pos,rect) ;

     if (_container->ChildCount(focus -> ViewID()) > 0) {
         UI_Point pt (_shape.Width()  - UI_Menu::SHADOW_WINDOW_SPACE - 
               UI_Menu::TRIANGLE_LFROMR - UI_Menu::TRIANGLE_WIDTH, 
               (UI_Menu::ITEM_VERT_SPACE / 2) + UI_Menu::TRIANGLE_TOP ) ;
         UI_ShadowedTriangle tempTri(
             UI_Rectangle(pt,UI_Menu::TRIANGLE_WIDTH,UI_Menu::TRIANGLE_HEIGHT),
             UI_ShadowedTriangle::East,UIRelief_Recessed ) ;
         tempTri.DrawOn(memDS);
     }

     memDS.CopyTo(sfc,UI_Point(UI_Menu::SHADOW_WINDOW_SPACE,
               focus -> _itemShape.Top() - (UI_Menu::ITEM_VERT_SPACE / 2)));

     DestroyDisplaySurface() ;

     if (_container->ChildCount(focus -> ViewID()) > 0) {
         UI_Point topLeft = focus -> _ItemWindowTopLeft() ;
         UI_Rectangle rect = focus -> _shape ;
         rect.Origin(topLeft) ;
         focus -> Shape() = rect ;
         XRaiseWindow(_Application -> AppDisplay(), focus -> ViewHandle());
         focus -> MakeVisible() ;
     }

     focus -> _NotifyGetFocus() ;

     return TRUE ;
}

bool UI_MenuItem::_ItemEraseShadowRect(UI_MenuItem* focus)
{
    if (!focus)
        return FALSE ;

    UI_Rectangle temp(UI_Menu::SHADOW_WINDOW_SPACE,
        focus -> _itemShape.Top() - (UI_Menu::ITEM_VERT_SPACE / 2),
        _shape.Width() - (2 * UI_Menu::SHADOW_WINDOW_SPACE),
        (focus -> _itemShape).Height() + UI_Menu::ITEM_VERT_SPACE) ;

    UI_DisplaySurface& sfc = CreateDisplaySurface() ;
    sfc.ColorRectangle (temp, _bgColor);

    CL_String label = focus -> Title() ;
    short mnemonic_pos = label.CharIndex (MNEMONIC_CHAR);
    if (mnemonic_pos >= 0) {
        label(mnemonic_pos,1) = "";
        if (label[mnemonic_pos] == MNEMONIC_CHAR)
            mnemonic_pos = -1 ;
    }
    
    sfc.WriteString (label, focus->_itemShape) ;
    _DrawToggleState (sfc, focus->_checkState, focus->_itemShape) ;
    
    if (mnemonic_pos >= 0 && mnemonic_pos < label.Size())
        focus->_DrawUnderLine (sfc, label, mnemonic_pos, focus->_itemShape);

     if (_container->ChildCount(focus -> ViewID()) > 0) {
         UI_Point pt (_shape.Width()  - UI_Menu::TRIANGLE_LFROMR - 
                      UI_Menu::TRIANGLE_WIDTH, 
                      focus -> _itemShape.Top() + UI_Menu::TRIANGLE_TOP ) ;
         UI_ShadowedTriangle tempTri(
             UI_Rectangle(pt,UI_Menu::TRIANGLE_WIDTH,UI_Menu::TRIANGLE_HEIGHT) ,
             UI_ShadowedTriangle::East,UIRelief_Raised) ;
         tempTri.DrawOn(sfc);
     }

    DestroyDisplaySurface() ;

    return TRUE ;

}

bool UI_MenuItem::_ItemLoseFocus()
{
    int child_count = _childCount ;

    if (!(IsVisible()) || child_count <= 0) {
        _NotifyLoseFocus() ;
        return TRUE ;
    }

    UI_MenuItem* child ;

    for (int i = 0; i < child_count; i++) {
        child = _container->Child(ViewID(),i);
        if (child -> IsVisible())
            child -> _ItemLoseFocus() ;
    }

    _focusItem = NULL ;
    MakeInvisible() ;
    _NotifyLoseFocus() ;
    return TRUE ;
}

void UI_MenuItem::_PrivateInitialize ()
{
    UI_SimpleVObject::_PrivateInitialize();
    if (!_enabled)
        Disable ();
}

UI_Point UI_MenuItem::_ItemWindowTopLeft()
{
    UI_MenuItem* parentItm = _container -> ParentItem(ViewID()) ;
    long xCoord = parentItm -> _shape.Left() + parentItm -> _shape.Width() ;
    long yCoord = _itemShape.Top() + parentItm -> _shape.Top() ;
    return UI_Point(xCoord,yCoord) ;
}


bool UI_MenuItem::DestroyVisualElement ()
{
}

bool UI_MenuItem::Enable ()
{
    _enabled = TRUE;
}



bool UI_MenuItem::Disable ()
{
    _enabled = FALSE;
}

bool UI_MenuItem::_TitleChanged ()
{
    _container -> _SetState() ;
    return TRUE ;
}

bool UI_MenuItem::_FontChanged ()
{
    // Override the inherited method with one that does nothing.
    return TRUE;
}


bool UI_MenuItem::PropagateFontChange (UI_VisualObject*)
{
    // Override the inherited method with one that does nothing.
    return TRUE;
}


UI_ViewHandle UI_MenuItem::ViewHandle () const
{
    return UI_ViewHandle(_window);
}


UI_MenuItem::~UI_MenuItem ()
{
}


bool UI_MenuItem::SetCheck (CheckState state)
{
    _checkState = state;
    return TRUE;
}

bool UI_MenuItem::ToggleCheckState ()
{
    if (_checkState == Checked)
        _checkState = NotChecked;
    else
        _checkState = Checked;
    return TRUE;
}


// --------------------- UI_Menu methods -----------------------


static const long ROOT_ID = 0;

UI_Menu::UI_Menu (UI_CompositeVObject* parent, UI_MenuItemDescriptor* items)
: UI_VisualObject (parent, UI_Rectangle (), -1)
{
    UI_MenuItem* root_item = new UI_MenuItem (this, "", ROOT_ID);
    if (!root_item)
        return;
    
    _isMouseDown = FALSE ;
    _isPtrGrabbed = FALSE ;

    _menuTree.NewRoot (ROOT_ID);
    _menuTree.Root()->SetContent((long) root_item);
    _BuildMenuSubtree (items, ROOT_ID);
    _UpdateChildCount() ;
    _focusItem = NULL;

    _font = new UI_Font (this);
    _ownFont = TRUE;
}


UI_ViewID UI_Menu::RootID () const
{
    return ROOT_ID;
}

short UI_Menu::_BuildMenuSubtree (UI_MenuItemDescriptor* items,
                                 UI_ViewID subtree_root)
{
    if (!items)
        return 0;
    CL_IntegerTreeNode* root = _menuTree.Node (subtree_root);
    if (!root)
        return 0;
    short n = 0;
    for (; items[n].label != NULL; n++) {
        UI_ViewID id;
        if (items[n].id < 0) {
            CL_Error::Warning ("UI_Menu::_Build: negative view id %d for "
                               "menu item", items[n].id);
            return n;
        }
        if (CL_String (UIMenu_Separator) == items[n].label)
            id = _AllocateSeparatorId();
        else
            id = items[n].id;
        UI_MenuItem* new_itm = new UI_MenuItem
            (this, items[n].label, id);
        CL_IntegerTreeNode* node = _menuTree.AddChild (id, subtree_root);
        if (node) {
            node->SetContent((long) new_itm);
            _BuildMenuSubtree (items[n].submenu, items[n].id);
        }
        else
            CL_Error::Warning ("YACL: UI_Menu: invalid (duplicate?) "
                               "menu item id %ld", id);
    }
    return n;
}

bool UI_Menu::Add (UI_ViewID id, const char* label,
                   UI_ViewID parent_id, short rank)
{
    if (id <= 0) {
        CL_Error::Warning ("UI_Menu::Add: negative view id %ld not allowed",
                           id);
        return FALSE; // Cannot add
    }
    CL_IntegerTreeNode* subtree_root = _menuTree.Node (parent_id);
    if (!subtree_root)
        return FALSE;
    UI_MenuItem* new_itm = new UI_MenuItem (this, label, id);
    CL_IntegerTreeNode* node = _menuTree.AddChild (id, parent_id, rank);
    if (node) {
        node->SetContent((long) new_itm);
        if (ViewHandle())
            _CreateMenuItemVisual (*node);
        _SetState() ;
        _UpdateChildCount() ;

        return TRUE;
    }
    return FALSE;
}



bool UI_Menu::AddSeparator (UI_ViewID parent_id, short rank)
{
    CL_IntegerTreeNode* subtree_root = _menuTree.Node (parent_id);
    if (!subtree_root)
        return FALSE;
    UI_ViewID id =  _AllocateSeparatorId();
    UI_MenuItem* new_itm = new UI_MenuItem (this, UIMenu_Separator, id);
    CL_IntegerTreeNode* node = _menuTree.AddChild (id, parent_id, rank);
    if (node) {
        node->SetContent((long) new_itm);
        if (ViewHandle())
            _CreateMenuItemVisual (*node);
        return TRUE;
    }

    _SetState() ;
    _UpdateChildCount() ;
    return FALSE;
}

bool UI_Menu::RemoveSeparator (UI_ViewID parent_id, short rank)
{
    CL_IntegerTreeNode* subtreeRoot = _menuTree.Node (parent_id);
    if (!subtreeRoot)
        return FALSE;
    short n = subtreeRoot->ChildCount();
    if (rank < 0 || rank >= n-1)
        return FALSE;
    CL_IntegerTreeNode* itemNode = subtreeRoot->Child (rank+1);
    if (!itemNode)
        return FALSE;
    return _DoRemove (itemNode);
}

bool UI_Menu::Remove (UI_ViewID id)
{
    CL_IntegerTreeNode* node = _menuTree.Node (id);
    if (!node)
        return FALSE;
    return _DoRemove (node);
}



short UI_Menu::ChildCount (UI_ViewID id) const
{
    CL_IntegerTreeNode* node = _menuTree.Node (id);
    return node ? node->ChildCount() : -1;
}


UI_MenuItem* UI_Menu::ParentItem(UI_ViewID id) const
{
    CL_IntegerTreeNode* node = _menuTree.Node (id);
    if (!node)
        return (UI_MenuItem*) NULL;
    node = node->Parent();
    return (node ? (UI_MenuItem*) node->Content () : (UI_MenuItem*) NULL);
}


UI_MenuItem* UI_Menu::Child (UI_ViewID id, short i) const
{
    CL_IntegerTreeNode* node = _menuTree.Node (id);
    if (!node)
        return (UI_MenuItem*) NULL;
    node = node->Child(i);
    return (node ? (UI_MenuItem*) node->Content () : (UI_MenuItem*) NULL);
}



short UI_Menu::Index (UI_ViewID id) const
{
    CL_IntegerTreeNode* node = _menuTree.Node (id);
    return node ? node->IndexInParent() : -1;
}

UI_MenuItem* UI_Menu::operator [] (UI_ViewID id)
{
    CL_IntegerTreeNode* node = _menuTree.Node (id);
    return node ? ((UI_MenuItem*) node->Content()) : (UI_MenuItem*) NULL;
}


UI_WindowClass UI_Menu::WindowClass () const
{
    // This method is not used for widget creation, only for run-time type
    // identification
    return X_YACL_MENU_CLASS;
}

bool UI_Menu::_CreateMenuItemVisual (CL_IntegerTreeNode& node)
{

     UI_MenuItem* item = (UI_MenuItem*) node.Content();

     if (!item || !node.Parent())
         return FALSE ;

     UI_MenuItem* parentItm = ParentItem(item -> ViewID());

     if (!(parentItm -> _window) || !(parentItm -> _visualElement))
     {
         //the node is a newly added leaf.

       UI_Rectangle pulldownMenuShape = _ItemWindowRectangle(*(node.Parent()));
       parentItm -> _SetShapeRectangle(pulldownMenuShape) ;
       Display* dp = YACLApp() -> AppDisplay() ;
       int screenNo = DefaultScreen(dp) ;
       parentItm->_visualElement = 
           new YVE_Window (parentItm,pulldownMenuShape,RootWindow(dp,screenNo));
       parentItm->_window = parentItm->_visualElement->Handle();
       parentItm -> MakeInvisible() ;
     }

     item -> _itemShape = _ItemShapeRectangle(node) ;

     if (!node.IsLeaf()) {
       UI_Rectangle pulldownMenuShape = _ItemWindowRectangle(node);
       item -> _SetShapeRectangle(pulldownMenuShape) ;
       Display* dp = YACLApp() -> AppDisplay() ;
       int screenNo = DefaultScreen(dp) ;
       item -> _visualElement = 
           new YVE_Window (item,pulldownMenuShape,RootWindow(dp,screenNo));
       item->_window = item->_visualElement->Handle();
     }

     item -> MakeInvisible() ;

     return (item -> _visualElement) && (item -> _window) ? TRUE : FALSE ;
}

bool UI_Menu::_GrabMouse()
{
    ulong mask = ButtonPressMask | ButtonReleaseMask | ButtonMotionMask
               | PointerMotionMask ;
    if (XGrabPointer (_Controller->AppDisplay(), ViewHandle(), False, 
           mask, GrabModeAsync, GrabModeAsync, None, 
           XCreateFontCursor(_Application -> AppDisplay(), XC_arrow), 
           CurrentTime) != GrabSuccess)
         return FALSE ; 

    _isPtrGrabbed = TRUE ;
    return TRUE ;
}

UI_MenuItem* UI_Menu::_MenuItemUnderMouse(int* wx, int* wy)
{

    Display* dp = _Controller -> AppDisplay() ;
    int screenNo = DefaultScreen(dp) ;

    int rx, ry;
    Window rw, cw;
    unsigned int mr;

    Window RW = RootWindow(dp, screenNo) ;

    if (XQueryPointer(_Controller -> AppDisplay(), RW,
                      &rw, &cw, &rx, &ry, wx, wy, &mr) == FALSE)
        return NULL ;

    int x, y ;

    if (cw == None)
        return NULL ;

    XTranslateCoordinates(dp, RW, cw, *wx, *wy, &x, &y, &rw) ;
    *wx = x ;
    *wy = y ;
    return ( (UI_MenuItem*) (*_Controller)[cw] ) ;

}

bool UI_Menu::_UngrabMouse()
{
    XUngrabPointer(_Application -> AppDisplay(),CurrentTime);
    return TRUE ;
}

void UI_Menu::_SetState()
{
    if (!_menuTree.Root())
        return ;
    CL_IntegerTreePreWalker walk (_menuTree.Root());
    CL_IntegerTreeNode* node ;
    while (walk.More()) {
        node = walk.Next();
        _ChangeMenuItem(*node);
    }
}

void UI_Menu::_UpdateChildCount()
{
    if (!_menuTree.Root())
        return ;
    CL_IntegerTreePreWalker walk (_menuTree.Root());
    CL_IntegerTreeNode* node ;
    UI_MenuItem* item ;
    while (walk.More()) {
        node = walk.Next();
        item = (UI_MenuItem*)node->Content() ;
        if (item)
            item -> _childCount = node->ChildCount() ;
    }
}

void UI_Menu::_ChangeMenuItem(CL_IntegerTreeNode& node)
{
     UI_MenuItem* item = (UI_MenuItem*) node.Content();

     if (!item) 
         return ;

     if ( (node.Parent() != _menuTree.Root()) && 
          (!node.Parent()) ) //This is the dummy root 
         return ;

     item -> _itemShape = _ItemShapeRectangle(node) ;

     if (!node.IsLeaf()) 
       item -> Shape() = _ItemWindowRectangle(node);
}


UI_Rectangle UI_Menu::_ItemShapeRectangle(CL_IntegerTreeNode& node) 
{
     UI_MenuItem* item = (UI_MenuItem*) node.Content();

     int xCoord,yCoord,width,height = _charHeight;
 
     if (!node.Parent()) 
         return Shape() ;

     if  (node.Parent() == _menuTree.Root()) {
         UI_Point p = this -> _ItemShapeTopLeft(node);
         xCoord = p.XCoord() ;
         yCoord = p.YCoord() ;
     }

     else {

         CL_IntegerTreeNode* parentNode = node.Parent();

         if (node.IndexInParent() > 0) {
             CL_IntegerTreeNode* sibling = 
                 parentNode -> Child(node.IndexInParent() - 1) ;

             UI_MenuItem* tmpItem = (UI_MenuItem*) (sibling -> Content()) ;
             xCoord = tmpItem -> _itemShape.Left() ;
             yCoord = tmpItem -> _itemShape.Top()    + 
                      tmpItem -> _itemShape.Height() + 
                      UI_Menu::ITEM_VERT_SPACE ;
         }
         else {
             if (item -> _checkState == UI_MenuItem::NotCheckable)
                 xCoord =  UI_Menu::ITEM_LEFT_SPACE ;
             else 
                 xCoord =  2 * UI_Menu::ITEM_LEFT_SPACE ;

             yCoord = (UI_Menu::ITEM_VERT_SPACE / 2) + 2 /*shadow*/ ; 
         }
     }

     CL_IntegerTreeNode* parentNode = node.Parent();
     CL_IntegerTreeNode* sibling = NULL ;
     if (parentNode)
         sibling = parentNode -> Child(node.IndexInParent() - 1) ;
     if (sibling)
     {
         UI_MenuItem* tmpItem = (UI_MenuItem*) (sibling -> Content()) ;
         if (tmpItem -> ViewID() >= 32000)
             yCoord -= (UI_Menu::ITEM_VERT_SPACE - UI_Menu::SEPARATOR_VERT_SPACE) ;
     }

     UI_MenuItem* sepItem = (UI_MenuItem*)node.Content() ;
     if (sepItem -> ViewID() >= 32000) {
         height = SEPARATOR_HEIGHT ;
         UI_MenuItem* sepParent = NULL ;
         sepParent = (UI_MenuItem*)node.Parent()->Content() ;
         yCoord -= (UI_Menu::ITEM_VERT_SPACE - UI_Menu::SEPARATOR_VERT_SPACE) ;
         if (sepParent)
             width = sepParent -> _shape.Width() - (2 * SEPARATOR_LSPACE) ;
     }
 
     else {
         UI_Font* font = OurFont() ;
         width = font->TextWidth(item->Title()); 
     }

     return UI_Rectangle(xCoord,yCoord,width,height) ;
}

//These two are the virtual methods.
//MenuBar and PopupMenu should override these.

UI_Point UI_Menu::_ItemShapeTopLeft(CL_IntegerTreeNode& node)
{
    return UI_Point();
}

UI_Point UI_Menu::_ItemWindowTopLeft(CL_IntegerTreeNode&)
{
    return UI_Point();
}

UI_Point UI_Menu::_ItemWindowTopLeft(UI_MenuItem*)
{
    return UI_Point();
}

UI_Rectangle UI_Menu::_ItemWindowRectangle(CL_IntegerTreeNode& node) 
{

    if (!node.Parent())
        return Shape() ;

    if (node.IsLeaf())
        return UI_Rectangle();

    int xCoord, yCoord, width = 0, height = _charHeight ;

    UI_Rectangle pulldownMenuShape ;

    UI_MenuItem* item = (UI_MenuItem*) node.Content();

    height = node.ChildCount() * (_charHeight + ITEM_VERT_SPACE);

    if (node.Parent() == _menuTree.Root()) {
      UI_Point p = this -> _ItemWindowTopLeft(node);
      xCoord = p.XCoord() ;
      yCoord = p.YCoord() ;
    }

    else {
        UI_MenuItem* parentItm = (UI_MenuItem*)(node.Parent()->Content()) ;
        xCoord = parentItm -> _shape.Left() + parentItm -> _shape.Width() ;
        yCoord = item -> _itemShape.Top() + parentItm -> Shape().Top() ;
    }

    height += WINDOW_XTRA_HT ;

    int sepItemHtConsumed = 0 ;
    bool isChildCheckable = FALSE ;

    int childCount = node.ChildCount() ;

    UI_MenuItem* child ;
    UI_Font* font = OurFont() ;

    for (int i = 0; i < childCount; i++)
    {
        child = Child(item -> ViewID(),i) ;

        if (child -> ViewID() >= 32000)
        sepItemHtConsumed += 
            (_charHeight - SEPARATOR_HEIGHT + 
            2 * (ITEM_VERT_SPACE - SEPARATOR_VERT_SPACE)) ; 

        if (child -> _checkState != UI_MenuItem::NotCheckable)
            isChildCheckable = TRUE ;

        int tmpWidth = font->TextWidth(child -> Title()) ; 

        if (width < tmpWidth)
            width = tmpWidth ;
    }

    width += (ITEM_LEFT_SPACE + ITEM_RIGHT_SPACE) ;
    height -= sepItemHtConsumed ;

    if (isChildCheckable)
        width += ITEM_LEFT_SPACE ;
   
    pulldownMenuShape = UI_Rectangle (xCoord,yCoord,width,height) ;

    return pulldownMenuShape ;

}

bool UI_Menu::HandleEvent (UI_Event* e)
{
    if (e->Type() == Event_LoseFocus) {
        _focusItem -> _NotifyLoseFocus() ;
        _focusItem = NULL;
    }

    return ProcessEvent (e);
}

void UI_Menu::_Processed()
{ 
    _isMouseDown = FALSE ;
    if (_isPtrGrabbed) {
        _UngrabMouse() ;
        _isPtrGrabbed = FALSE ;
    }

    if (_focusItem) {
        _focusItem -> _ItemLoseFocus() ;
        _focusItem = NULL ;
    }

    if (CL_String (ClassName()) == "UI_PopupMenu" && IsVisible() ) {
        MakeInvisible();
    }
}

void UI_Menu::_ButtonDown(bool press)
{
    _isMouseDown = press; 
}

bool UI_Menu::_isButtonDown()
{
    return _isMouseDown;
}

UI_MenuItem* UI_Menu::_CurrentFocusItem(UI_Point p)
{        
    if (Shape().HitTest(p) != UIHit_Inside)
        return NULL ;

    int childCount = _menuTree.Root() -> ChildCount() ;

    UI_MenuItem* itm ; 

    for (int i = 0; i < childCount; i++) {
        itm = (UI_MenuItem*) (_menuTree.Root()->Child(i)->Content());
        if ( (itm->_itemShape).HitTest(p) == UIHit_Inside)
            return itm ;
    }

    return NULL ;
}

bool UI_Menu::_DrawShadowRect(UI_DrawingSurface& sfc)
{
    if (_focusItem) {
        UI_ShadowRectangle temp((_focusItem -> _itemShape).Left() - 
                                 SHADOW_ITEM_SPACE, 
             (_focusItem -> _itemShape).Top() - SHADOW_ITEM_TOP_SPACE, 
            _focusItem->_itemShape.Width() + (2 * SHADOW_ITEM_SPACE) ,
             _focusItem->_itemShape.Height() +  (2 * SHADOW_ITEM_TOP_SPACE) ,
             UI_ShadowRectangle::Raised, 2, FALSE, _bgColor);

        temp.DrawOn(sfc);

        CL_String label = _focusItem -> Title() ;
            short mnemonic_pos = label.CharIndex (MNEMONIC_CHAR);
        if (mnemonic_pos >= 0) {
            label(mnemonic_pos,1) = "";
            if (label[mnemonic_pos] == MNEMONIC_CHAR)
                mnemonic_pos = -1 ;
        }

        sfc.WriteString(label,_focusItem->_itemShape); 
        if (mnemonic_pos >= 0 && mnemonic_pos < label.Size())
            _focusItem -> _DrawUnderLine(sfc,label,
              mnemonic_pos,_focusItem->_itemShape) ;
        return TRUE ;
    }

    return FALSE ;
}

bool UI_Menu::_EraseShadowRect()
{
    if (_focusItem) {
        UI_DisplaySurface& sfc = CreateDisplaySurface() ;
        UI_Rectangle temp(
            _focusItem->_itemShape.Left()   - SHADOW_ITEM_SPACE,
            _focusItem->_itemShape.Top()    - SHADOW_ITEM_TOP_SPACE,
            _focusItem->_itemShape.Width()  + (2 * SHADOW_ITEM_SPACE),
            _focusItem->_itemShape.Height() + (2 * SHADOW_ITEM_TOP_SPACE)); 

        sfc.ColorRectangle(temp,_bgColor);

        CL_String label = _focusItem -> Title() ;
            short mnemonic_pos = label.CharIndex (MNEMONIC_CHAR);
        if (mnemonic_pos >= 0)
            label(mnemonic_pos,1) = "";

        sfc.WriteString(label , _focusItem->_itemShape);
        DestroyDisplaySurface() ;
        return TRUE ;
    }

    return FALSE ;

}
    
bool UI_Menu::MakeVisualElement ()
{
    return UI_VisualObject::MakeVisualElement ();
}

bool UI_Menu::DestroyVisualElement ()
{
    return FALSE;
}


UI_Menu::~UI_Menu()
{
}


bool UI_Menu::_FontChanged ()
{
    // Override the inherited method with one that does nothing.
    return TRUE;
}


bool UI_Menu::PropagateFontChange (UI_VisualObject*)
{
    // Override the inherited method with one that does nothing.
    return TRUE;
}


bool UI_Menu::_DoRemove (CL_IntegerTreeNode* node)
{
    if (!node)
        return FALSE;
    _focusItem = NULL;
    UI_MenuItem* itm = (UI_MenuItem*) node->Content();
    if (!itm)
        return FALSE;
    UI_ViewID id = itm ->ViewID();
    _menuTree.DestroySubtree (id);
    _SetState() ;
    _UpdateChildCount() ;
    return TRUE;
}


UI_ViewID UI_Menu::_AllocateSeparatorId()
{
    // Look through the menu tree for the first unused id above 32000, and
    // return it.
    long i = 32000;
    while (_menuTree.Node(i))
        i++;
    return i;
}



// ----------------- UI_MenuBar methods -------------------------

UI_MenuBar::UI_MenuBar (UI_CompositeVObject* parent,
                        UI_MenuItemDescriptor* item)
: UI_Menu (parent, item)
{

    MenuBinding1 bind(this,&UI_MenuBar::_MenuReconfigure); 
    parent -> AddEventDependent(Event_Reconfigure,bind); 

    parent->SetMenuBar (this);

}

UI_MenuBar::~UI_MenuBar()
{
    MenuBinding1 bind(this,&UI_MenuBar::_MenuReconfigure); 
    Parent() -> RemoveEventDependent(Event_Reconfigure,bind) ;
}

bool UI_MenuBar::MakeVisualElement ()
{
    _SetShapeRectangle (UI_Rectangle (0, 0, _parent->Shape().Width() , 
                                      MENU_BAR_HEIGHT));

    UI_MenuItem* root_item = (UI_MenuItem*)
        _menuTree.Node (ROOT_ID)->Content();

    UI_Rectangle temp(0,0,_parent->Shape().Width(),MENU_BAR_HEIGHT);

    root_item->_visualElement = new YVE_Window (root_item,temp);
    root_item->_window = root_item->_visualElement->Handle();

    _visualElement = root_item->_visualElement ;
    _window = root_item->_window;


    CL_IntegerTreePreWalker walk (_menuTree.Root());

    while (walk.More()) {
        CL_IntegerTreeNode* node = walk.Next();
        _CreateMenuItemVisual(*node);
        UI_MenuItem* tmp = (UI_MenuItem*)node->Content() ;
    }
    _created = TRUE;

    return TRUE;
}

bool UI_MenuBar::Add (UI_ViewID id, const char* label,
                      UI_ViewID parent_id, short rank)
{
    bool b = UI_Menu::Add (id, label, parent_id, rank);
    return b;
}

UI_Point UI_MenuBar::_ItemShapeTopLeft(CL_IntegerTreeNode& node)
{

    int xCoord = 0, yCoord = 0; 

    if (node.IndexInParent() > 0)
    {
        CL_IntegerTreeNode* parentNode = node.Parent();
        CL_IntegerTreeNode* sibling = 
            parentNode -> Child(node.IndexInParent() - 1) ;
        UI_MenuItem* tmpItem = (UI_MenuItem*) (sibling -> Content()) ;

        xCoord = tmpItem -> _itemShape.Left() +
                 tmpItem -> _itemShape.Width() + 
                 ITEM_LEFT_SPACE ;
        yCoord = MENU_ITEM_TOP_SPACE ;

    }
    else {
            xCoord = ITEM_LEFT_SPACE ;
            yCoord = MENU_ITEM_TOP_SPACE ;
    }

    return UI_Point(xCoord,yCoord) ;
}

UI_Point UI_MenuBar::_ItemWindowTopLeft(CL_IntegerTreeNode& node)
{
    int xCoord = 0, yCoord = 0;

    UI_MenuItem* item = (UI_MenuItem*)node.Content() ;

    if (!item)
        return UI_Point() ;

    Display* dp = _Application -> AppDisplay() ;
    int screenNo = DefaultScreen(dp) ;

    Window rw =  RootWindow(dp,screenNo)  ;
    Window pw =  Parent() -> ViewHandle() ;

    int x_ret,y_ret ;
    Window dummy;

    XTranslateCoordinates(dp,pw,rw,0,0,&x_ret,&y_ret,&dummy) ;

    xCoord = item -> _itemShape.Left() - SHADOW_ITEM_SPACE + x_ret ;
    yCoord = item -> _itemShape.Top() + item -> _itemShape.Height() + 
             y_ret + SHADOW_ITEM_TOP_SPACE ;

    return UI_Point(xCoord,yCoord) ;
}

UI_Point UI_MenuBar::_ItemWindowTopLeft(UI_MenuItem* item)
{
    int xCoord = 0, yCoord = 0;

    if (!item)
        return UI_Point() ;

    Display* dp = _Application -> AppDisplay() ;
    int screenNo = DefaultScreen(dp) ;

    Window rw =  RootWindow(dp,screenNo)  ;
    Window pw =  Parent() -> ViewHandle() ;

    int x_ret,y_ret ;
    Window dummy;

    XTranslateCoordinates(dp,pw,rw,0,0,&x_ret,&y_ret,&dummy) ;

    xCoord = item -> _itemShape.Left() - SHADOW_ITEM_SPACE + x_ret ;
    yCoord = item -> _itemShape.Top() + item -> _itemShape.Height() + 
             y_ret + SHADOW_ITEM_TOP_SPACE ;

    return UI_Point(xCoord,yCoord) ;
}

bool UI_MenuBar::_MenuReconfigure(UI_Event& evt)
{
    UI_Rectangle rect (0,0,
          _parent->Shape().Width(),MENU_BAR_HEIGHT) ;

    Shape() = rect;

    UI_MenuItem* root_item = (UI_MenuItem*)
        _menuTree.Node (ROOT_ID)->Content();

    root_item -> Shape() = rect;

    return TRUE;
}

bool UI_MenuBar::HandleEvent(UI_Event* e)
{
    switch(e->Type()) {

    case Event_Paint: {
    
        UI_DwgSurfaceInMemory memDS (Shape().Width(),
                                 Shape().Height());
        memDS.ColorRectangle(memDS.DrawingArea(),_bgColor);

        UI_ShadowRectangle temp(0,0,Shape().Width(),Shape().Height(),
                            UI_ShadowRectangle::Raised,2,FALSE,_bgColor);
    
        temp.DrawOn(memDS);

        int childCount = _menuTree.Root() -> ChildCount() ;

        UI_MenuItem* itm ; 
        for (int i = 0; i < childCount; i++) {
            itm = (UI_MenuItem*) (_menuTree.Root() -> Child(i) -> Content());
            CL_String label = itm -> Title() ;
            short mnemonic_pos = label.CharIndex (MNEMONIC_CHAR);
            if (mnemonic_pos >= 0 && mnemonic_pos < label.Size()) {
                label(mnemonic_pos,1) = "";
                if (label[mnemonic_pos] == MNEMONIC_CHAR)
                    mnemonic_pos = -1 ;
            }

            UI_Color memDSColor ;

            if (!(itm -> _enabled)) {
                memDSColor = memDS.Pen().Color() ;
                memDS.Pen().Color(DisabledForeground()) ;
            }

            memDS.WriteString(label,itm->_itemShape);
        
            if (mnemonic_pos >= 0 && mnemonic_pos < label.Size())
                itm -> _DrawUnderLine( memDS, label, 
                    mnemonic_pos,itm->_itemShape ) ;
            
            memDS.Pen().Color(memDSColor);

        }

        _DrawShadowRect(memDS) ;

        UI_DisplaySurface& sfc = CreateDisplaySurface();
        memDS.CopyTo(sfc);
        DestroyDisplaySurface();
         
        return TRUE ; 
    }

    case Event_LButtonPress: 
    case Event_RButtonPress: 
    {
        if (_isPtrGrabbed) {
            _isMouseDown = TRUE ;

            if (_shape.HitTest(e->Position()) == UIHit_Inside) {
                // Clicked on menubar
                UI_MenuItem* currMenuBarItem = _CurrentFocusItem(e->Position());
                if (!currMenuBarItem) {
                    _Processed() ; 
                    return FALSE ;
                }

                if (!(currMenuBarItem -> _enabled)) 
                    return TRUE ; 
                if (_focusItem) {
                    if (currMenuBarItem -> ViewID() == _focusItem -> ViewID()) {
                         _isMouseDown = TRUE ;
                         return TRUE ;
                    }
                    _EraseShadowRect() ;
                    _focusItem -> _ItemLoseFocus() ;
                }

                UI_Point topLeft = 
                   this -> _ItemWindowTopLeft(currMenuBarItem) ;
                UI_Rectangle rect = currMenuBarItem -> Shape() ;
                rect.Origin(topLeft) ;

                currMenuBarItem -> _NotifyGetFocus() ;

                currMenuBarItem -> Shape() = rect ;
                XRaiseWindow( _Controller -> AppDisplay(),
                         currMenuBarItem -> ViewHandle() ) ;
                currMenuBarItem -> MakeVisible() ;
                _focusItem = currMenuBarItem ;
    
                UI_DisplaySurface& sfc = CreateDisplaySurface() ;
                _DrawShadowRect(sfc) ;
                DestroyDisplaySurface() ;

                return TRUE ;
            }

            int x, y ;
            UI_MenuItem* currItem = _MenuItemUnderMouse(&x,&y); 

            if (currItem) { //Notify the button press for the approp. menuitem
                currItem -> _ButtonPress(x,y) ;
                return TRUE ;
            }
            _Processed() ;
            return FALSE ;
        }

        //Click on menubar for the first time

        UI_MenuItem* currFocus = _CurrentFocusItem(e->Position()) ;
 
        if (currFocus)
        {
            if (!(currFocus -> _enabled)) 
                return TRUE ; 

            _GrabMouse() ;
            UI_Point topLeft = 
               this -> _ItemWindowTopLeft(currFocus) ;
            UI_Rectangle rect = currFocus -> Shape() ;
            rect.Origin(topLeft) ;

            currFocus -> _NotifyGetFocus() ;

            currFocus -> Shape() = rect ;
            XRaiseWindow( _Controller -> AppDisplay(),
                     currFocus -> ViewHandle() ) ;
            currFocus -> MakeVisible() ;
            _isMouseDown = TRUE ;
            _focusItem = currFocus ;

            UI_DisplaySurface& sfc = CreateDisplaySurface() ;
            _DrawShadowRect(sfc) ;
            DestroyDisplaySurface() ;
            return TRUE ; 

        }
        return FALSE ; 
    }

    case Event_MouseMove: {

        if (!_isMouseDown || !_isPtrGrabbed) 
            return FALSE ;

        if (_shape.HitTest(e->Position()) == UIHit_Inside) {
            // Mouse move on menubar
            UI_MenuItem* currMenuBarItem = _CurrentFocusItem(e->Position());
 
            if (!currMenuBarItem) 
                return TRUE ;
            if (!(currMenuBarItem -> _enabled)) 
                return TRUE ;

            if (_focusItem) {
                if (currMenuBarItem -> ViewID() == _focusItem -> ViewID())
                    return TRUE ;
                _EraseShadowRect() ;
                _focusItem -> _ItemLoseFocus() ;
            }

            UI_Point topLeft = 
                this -> _ItemWindowTopLeft(currMenuBarItem) ;
            UI_Rectangle rect = currMenuBarItem -> Shape() ;
            rect.Origin(topLeft) ;

            currMenuBarItem -> _NotifyGetFocus() ;

            currMenuBarItem -> Shape() = rect ;
            XRaiseWindow( _Controller -> AppDisplay(),
                          currMenuBarItem -> ViewHandle() ) ;
            currMenuBarItem -> MakeVisible() ;
            _focusItem = currMenuBarItem ;

            UI_DisplaySurface& sfc = CreateDisplaySurface() ;
            _DrawShadowRect(sfc) ;
            DestroyDisplaySurface() ;
            return TRUE ;

        }

        int x, y ;
        UI_MenuItem* currItem = _MenuItemUnderMouse(&x,&y); 

        if (currItem) { //Notify the mouse move for the approp. menuitem
            currItem -> _MouseMove(x,y) ;
            return TRUE ;
        }
        return FALSE ;
    }

    case Event_LButtonRelease: 
    case Event_RButtonRelease: 
    {
        if (!_isMouseDown || !_isPtrGrabbed) {
            _isMouseDown = FALSE ;
            return FALSE ;
        }

        _isMouseDown = FALSE ;

        if (_shape.HitTest(e->Position()) == UIHit_Inside) {
            UI_MenuItem* currMenuBarItem = _CurrentFocusItem(e->Position());
            if (!currMenuBarItem) {
                _Processed() ; 
                return FALSE ;
            }

            currMenuBarItem -> _NotifyLoseFocus() ;

            if (ChildCount(currMenuBarItem -> ViewID()) <= 0) {
                _Processed() ; 
                UI_Event evt(Event_Select,currMenuBarItem,currMenuBarItem) ;
                _Controller -> DispatchEvent(&evt) ;
            }
            return TRUE ;
        }

        int x, y ;
        UI_MenuItem* currItem = _MenuItemUnderMouse(&x,&y); 

        if (currItem) { //Notify the button release for the approp. menuitem
            currItem -> _ButtonRelease(x,y) ;
            return TRUE ;
        }
        _Processed() ;
        return FALSE ;
    }

    default:
       return ProcessEvent(e) ;

    }

    return ProcessEvent(e) ;

}

UI_WindowClass UI_MenuBar::WindowClass () const
{
    // This method is not used for widget creation, only for run-time type
    // identification
    return X_YACL_MENUBAR_CLASS;
}

// ---------------------------- UI_PopupMenu methods ---------------------

UI_PopupMenu::UI_PopupMenu (UI_CompositeVObject* parent,
                            UI_MenuItemDescriptor* item)
: UI_Menu (parent, item)
{
    _visible = FALSE; // Initially invisible
}


UI_PopupMenu::~UI_PopupMenu ()
{
}

UI_Point UI_PopupMenu::_ItemShapeTopLeft(CL_IntegerTreeNode& node)
{

    int xCoord = 0, yCoord = 0 ;

    CL_IntegerTreeNode* parentNode = node.Parent();

    if (!parentNode)
        return UI_Point() ;

    UI_MenuItem* item = (UI_MenuItem*)node.Content() ;

    if (node.IndexInParent() > 0) {
        CL_IntegerTreeNode* sibling = 
            parentNode -> Child(node.IndexInParent() - 1) ;
        UI_MenuItem* tmpItem = (UI_MenuItem*) (sibling -> Content()) ;

        xCoord = tmpItem -> _itemShape.Left() ;
        yCoord = tmpItem -> _itemShape.Top()    + 
                 tmpItem -> _itemShape.Height() + 
                 ITEM_VERT_SPACE ;
    }
    else {
         if (item -> _checkState == UI_MenuItem::NotCheckable)
             xCoord =  ITEM_LEFT_SPACE ;
         else 
             xCoord =  2 * ITEM_LEFT_SPACE ;

         yCoord = (ITEM_VERT_SPACE / 2) + 2 /*shadow*/ ; 
    }

    return UI_Point(xCoord,yCoord) ;

}

UI_Point UI_PopupMenu::_ItemWindowTopLeft(CL_IntegerTreeNode& node)
{

    UI_MenuItem* item = (UI_MenuItem*)(node.Content()) ;
    UI_MenuItem* parentItm = (UI_MenuItem*)(node.Parent()->Content()) ;

    int xCoord = 
        (parentItm -> Shape()).Left() + (parentItm -> Shape()).Width() ;
    int yCoord = 
        (item -> _itemShape).Top() + (parentItm -> Shape()).Top() ;

    return UI_Point(xCoord,yCoord) ;
}

UI_Point UI_PopupMenu::_ItemWindowTopLeft(UI_MenuItem* item)
{

    if (!item)
        return UI_Point() ;

    UI_MenuItem* parentItm = ParentItem(item -> ViewID());

    int xCoord = 
        (parentItm -> Shape()).Left() + (parentItm -> Shape()).Width() ;
    int yCoord = 
        (item -> _itemShape).Top() + (parentItm -> Shape()).Top() ;

    return UI_Point(xCoord,yCoord) ;
}

bool UI_PopupMenu::MakeVisualElement ()
{
    UI_MenuItem* root_item = (UI_MenuItem*)
        _menuTree.Node (ROOT_ID)->Content();

    int childCount = _menuTree.Root() -> ChildCount() ;
    int width = 0 , sepItemHtConsumed = 0 , height = 0 ;

    UI_MenuItem* child ; 

    UI_Font* font = OurFont() ;

    for (int i = 0; i < childCount; i++) {
        child = (UI_MenuItem*) (_menuTree.Root() -> Child(i) -> Content());

        if (!child)
            return FALSE ;

        int tmpWidth = font->TextWidth(child -> Title());
        width = (width < tmpWidth) ? tmpWidth : width ;

        if (child -> ViewID() >= 32000)
            sepItemHtConsumed += 
                (_charHeight - SEPARATOR_HEIGHT + 
                2 * (ITEM_VERT_SPACE - SEPARATOR_VERT_SPACE)) ; 
    }

    height = childCount * (_charHeight + ITEM_VERT_SPACE);
    height += WINDOW_XTRA_HT ;
    width += (ITEM_LEFT_SPACE + ITEM_RIGHT_SPACE) ;
        
    _SetShapeRectangle (UI_Rectangle (0, 0, width, height));

    UI_Rectangle temp(0,0,width, height);

    Display* dp = YACLApp() -> AppDisplay() ;
    int screenNo = DefaultScreen(dp) ;

    
    root_item->_visualElement = 
        new YVE_Window (root_item, temp, RootWindow(dp,screenNo));
    root_item->_window = root_item->_visualElement->Handle();

    _visualElement = root_item->_visualElement ;
    _window = root_item->_window;
    root_item -> MakeInvisible() ;
    MakeInvisible() ;

    CL_IntegerTreePreWalker walk (_menuTree.Root());
    CL_IntegerTreeNode* node ;
    while (walk.More()) {
        node = walk.Next();
        _CreateMenuItemVisual (*node);
    }
    _created = TRUE;
    return TRUE;
}

bool UI_PopupMenu::ShowAt (const UI_Point& p)
{
    _isMouseDown = TRUE ;
    _currentPosition = p ;
    _SetState(p) ;

    XRaiseWindow(_Application -> AppDisplay(),ViewHandle());
    MakeVisible() ;
    return TRUE ;
}

bool UI_PopupMenu::HandleEvent (UI_Event* e)
{

    switch(e -> Type()) {

    case Event_Paint: {
        _GrabMouse() ;
        UI_MenuItem* root_item = 
             (UI_MenuItem*)(_menuTree.Root() -> Content()) ;
        _focusItem = NULL ;
        _isMouseDown = TRUE ;

        UI_DwgSurfaceInMemory memDS (Shape().Width(),
                                 Shape().Height());
        memDS.ColorRectangle(memDS.DrawingArea(),_bgColor);
        UI_ShadowRectangle temp(0,0,Shape().Width(),Shape().Height(),
                            UI_ShadowRectangle::Raised,2,FALSE,_bgColor);
        temp.DrawOn(memDS);

        int childCount = _menuTree.Root() -> ChildCount() ;
        UI_MenuItem* itm ; 
        for (int i = 0; i < childCount; i++) {
            itm = (UI_MenuItem*) (_menuTree.Root() -> Child(i) -> Content());
            if (itm -> ViewID() >= 32000) {
                UI_ShadowRectangle temp(SEPARATOR_LSPACE,
                    itm -> _itemShape.Top(),
                    itm -> _itemShape.Width(), 
                    itm -> _itemShape.Height(),
                    UI_ShadowRectangle::Recessed,1,FALSE,_bgColor);
                temp.DrawOn(memDS);
            }
            else {
                CL_String label = itm -> Title() ;
                short mnemonic_pos = label.CharIndex (MNEMONIC_CHAR);
                if (mnemonic_pos >= 0 && mnemonic_pos < label.Size()) {
                    label(mnemonic_pos,1) = "";
                    if (label[mnemonic_pos] == MNEMONIC_CHAR)
                        mnemonic_pos = -1 ;
                }
                UI_Color memDSColor = memDS.Pen().Color() ;
                if (!(itm -> _enabled))
                    memDS.Pen().Color(DisabledForeground()) ;
                UI_Rectangle rect( (itm -> _itemShape).Left(),
                                   (itm -> _itemShape).Top(),
                                   (itm -> _itemShape).Width(),
                                   (itm -> _itemShape).Height() ) ;

                memDS.WriteString(label,rect);

                if (ChildCount(itm -> ViewID()) > 0) {
                    UI_Point pt (Shape().Width()  - TRIANGLE_LFROMR 
                                 - TRIANGLE_WIDTH, 
                                 (itm -> _itemShape).Top() + TRIANGLE_TOP ) ;
                    UI_ShadowedTriangle tempTri(
                        UI_Rectangle(pt,TRIANGLE_WIDTH,TRIANGLE_HEIGHT) ,
                        UI_ShadowedTriangle::East,UIRelief_Raised ) ;
                    tempTri.DrawOn(memDS);
                }

                if (mnemonic_pos >= 0 && mnemonic_pos < label.Size())
                    itm -> _DrawUnderLine( memDS, label, 
                        mnemonic_pos,itm->_itemShape ) ;
            
                memDS.Pen().Color(memDSColor) ;
            }
        }

        UI_DisplaySurface& sfc = CreateDisplaySurface();
        memDS.CopyTo(sfc);
        DestroyDisplaySurface();

        return TRUE ; 
    }

    case Event_MouseMove: {

        Display* dp = _Application -> AppDisplay() ;
        int screenNo = DefaultScreen(dp) ;

        int rx, ry, wx, wy;
        Window rw, cw;
        unsigned int mr;

        Window RW = RootWindow(dp, screenNo) ;

        if (XQueryPointer(_Controller -> AppDisplay(), RW,
                          &rw, &cw, &rx, &ry, &wx, &wy, &mr) == FALSE)
            return FALSE ;

        if ( (!_isMouseDown) || (!_isPtrGrabbed) )
            return TRUE ;
         
        UI_Point p = e -> Position() ;

        if (cw == None) 
            return FALSE ;

        if (cw == _window) {
            UI_MenuItem* root_item = (UI_MenuItem*)(_menuTree.Root() -> Content()) ;       
            UI_MenuItem* currFocus = root_item -> _CurrentFocusItem(p);

            if (!currFocus)  {
                if (_focusItem) {
                    root_item -> _ItemEraseShadowRect(_focusItem) ;
                    _focusItem -> _ItemLoseFocus() ;
                    _focusItem = NULL ;
                }
                return TRUE;
            }

            if (!(currFocus -> _enabled)) {
                if (_focusItem) {
                    root_item -> _ItemEraseShadowRect(_focusItem) ;
                    _focusItem -> _ItemLoseFocus() ;
                    _focusItem = NULL ;
                }
                return TRUE ;
            }

            if (_focusItem) {
                if (currFocus -> ViewID() == _focusItem -> ViewID())
                    return TRUE ;
    
                root_item -> _ItemEraseShadowRect(_focusItem) ;
                _focusItem -> _ItemLoseFocus() ;
            }
            root_item -> _ItemGetFocus(currFocus) ;
            _focusItem = currFocus ;
            return TRUE ;
        }

        int x, y ;
        UI_MenuItem* currItem = ( (UI_MenuItem*) (*_Controller)[cw] ) ;

        if (currItem) { //Notify the mouse move for the approp. menuitem
            XTranslateCoordinates(dp,RW,cw,wx,wy,&x,&y,&rw) ;
            currItem -> _MouseMove(x,y) ;
            return TRUE ;
        }
        return FALSE ;
    }

    case Event_LButtonRelease:
    case Event_MButtonRelease:
    case Event_RButtonRelease: 
    {
        _isMouseDown = FALSE ;

        UI_Point p = e -> Position() ;

        Display* dp = _Application -> AppDisplay() ;
        int screenNo = DefaultScreen(dp) ;

        int rx, ry, wx, wy;
        Window rw, cw;
        unsigned int mr;

        Window RW = RootWindow(dp, screenNo) ;

        if (XQueryPointer(_Controller -> AppDisplay(), RW,
                          &rw, &cw, &rx, &ry, &wx, &wy, &mr) == FALSE)
            return FALSE ;

        if (cw == None) {
            _Processed() ;
            return FALSE ;
        }


        if (cw == _window) {
            UI_MenuItem* root_item = (UI_MenuItem*)(_menuTree.Root() -> Content()) ;       
            UI_MenuItem* currFocus = root_item -> _CurrentFocusItem(p);

            if (currFocus)  {

                if (ChildCount(currFocus -> ViewID()) > 0)
                    return TRUE ;

                if (! (currFocus -> _enabled) ) {
                    _Processed() ;
                    return TRUE ;
                }

                _Processed() ;
                UI_Event evt(Event_Select,currFocus,currFocus) ;
                _Controller -> DispatchEvent(&evt) ;
                return TRUE ;
            }

            _Processed() ;
            return FALSE ;
        }

        int x, y ;
        UI_MenuItem* currItem = ( (UI_MenuItem*) (*_Controller)[cw] ) ;

        if (currItem) { //Notify the button release for the approp. menuitem
            XTranslateCoordinates(dp,RW,cw,wx,wy,&x,&y,&rw) ;
            currItem -> _ButtonRelease(x,y) ;
            return TRUE ;
        }

        _Processed() ;
        return FALSE ;
    }

    case Event_LButtonPress:
    case Event_MButtonPress:
    case Event_RButtonPress: 
    {
        _isMouseDown = TRUE ;

        UI_Point p = e -> Position() ;

        Display* dp = _Application -> AppDisplay() ;
        int screenNo = DefaultScreen(dp) ;

        int rx, ry, wx, wy;
        Window rw, cw;
        unsigned int mr;

        Window RW = RootWindow(dp, screenNo) ;

        if (XQueryPointer(_Controller -> AppDisplay(), RW,
                          &rw, &cw, &rx, &ry, &wx, &wy, &mr) == FALSE)
            return FALSE ;

        if (cw == None) {
            _Processed() ;
            return FALSE ;
        }

        if (cw == _window) {
            UI_MenuItem* root_item = (UI_MenuItem*)(_menuTree.Root() -> Content()) ;       
            UI_MenuItem* currFocus = root_item -> _CurrentFocusItem(p);

            if (!currFocus) {
                if (_focusItem) {
                    root_item -> _ItemEraseShadowRect(_focusItem) ;
                    _focusItem -> _ItemLoseFocus() ;
                }
                return TRUE ;
            }

            if (_focusItem) {
                if (_focusItem -> ViewID() == currFocus -> ViewID())
                   return TRUE ;

                root_item -> _ItemEraseShadowRect(_focusItem) ;
                _focusItem -> _ItemLoseFocus() ;
            }

            if (! (currFocus -> _enabled) )
                return TRUE ;

            root_item -> _ItemGetFocus(currFocus) ;
            _focusItem = currFocus ;
            return TRUE ;
        }

        int x, y ;
        UI_MenuItem* currItem = ( (UI_MenuItem*) (*_Controller)[cw] ) ;

        if (currItem) { //Notify the button press for the approp. menuitem
            XTranslateCoordinates(dp,RW,cw,wx,wy,&x,&y,&rw) ;
            currItem -> _ButtonPress(x,y) ;
            return TRUE ;
        }

        _Processed() ;
        return FALSE ;

    }

    default: 
        return ProcessEvent(e) ;
    }

}


bool UI_PopupMenu::_SetState (UI_Point pt)
{
    Display* dp = _Application -> AppDisplay() ;
    int screenNo = DefaultScreen(dp) ;

    Window rw =  RootWindow(dp,screenNo)  ;
    Window pw =  Parent() -> ViewHandle() ;

    int x_ret,y_ret ;
    Window dummy;

    XTranslateCoordinates (dp, pw, rw, 0, 0, &x_ret, &y_ret, &dummy) ;

    UI_Point p = UI_Point (x_ret,y_ret) + pt ;

    UI_Rectangle rect(p,Shape().Width(),Shape().Height() );
    Shape() = rect ;
    UI_MenuItem* root_item = (UI_MenuItem*)_menuTree.Root() -> Content();
    root_item -> Shape() = rect ;
    return TRUE ;
}



UI_WindowClass UI_PopupMenu::WindowClass () const
{
    // This method is not used for widget creation, only for run-time type
    // identification
    return X_YACL_POPUPMENU_CLASS;
}

