
/*
 *
 *          Copyright (C) 1997, M. A. Sridhar
 *  
 *
 *     This software is Copyright M. A. Sridhar, 1997. You are free
 *     to copy, modify or distribute this software  as you see fit,
 *     and to use  it  for  any  purpose, provided   this copyright
 *     notice and the following   disclaimer are included  with all
 *     copies.
 *
 *                        DISCLAIMER
 *
 *     The author makes no warranties, either expressed or implied,
 *     with respect  to  this  software, its  quality, performance,
 *     merchantability, or fitness for any particular purpose. This
 *     software is distributed  AS IS.  The  user of this  software
 *     assumes all risks  as to its quality  and performance. In no
 *     event shall the author be liable for any direct, indirect or
 *     consequential damages, even if the  author has been  advised
 *     as to the possibility of such damages.
 *
 */



#include "ui/treeview.h"
#include "ui/applic.h"

#include "base/treewalk.h"
#include "base/genseq.h"
#include "base/strgseq.h"


#if defined(__OS2__)
#    define TREESTYLE  WS_VISIBLE
struct TreeItem {
    RECORDCORE core;
    long       nodeId;
};
static const long ExtraBytes = sizeof (TreeItem) - sizeof (RECORDCORE);

#elif defined(__MS_WIN32__)
#include <commctrl.h>
#endif



UI_TreeView::UI_TreeView (UI_VisualObject* parent, const UI_Rectangle& shape,
                          UI_ViewID id)
    : UI_SimpleVObject (parent, shape, id), _selectedNode (-1)
{
    _model      = new UI_TreeViewModel;
    _ownModel   = TRUE;
    _separation = -1;
    _indent     = -1;
#if defined(__OS2__)
    _style = TREESTYLE;
#endif
    ((UI_TreeViewModel*) _model)->AddClient (this);
}



UI_TreeView::UI_TreeView (UI_VisualObject* parent, const UI_Rectangle& shape,
                          UI_ViewID id, UI_TreeViewModel* model)
    : UI_SimpleVObject (parent, model, id, shape), _selectedNode (-1)
{
    _separation = -1;
    _indent     = -1;
    model->AddClient (this);
#if defined(__OS2__)
    _style = TREESTYLE;
#endif
}


UI_TreeView::~UI_TreeView ()
{
    _DoDestroy (((UI_TreeViewModel*) _model)->Root());
    // Note that we have to DoDestroy regardless of whether the model is
    // owned or not, because DoDestroy merely destroys the window system's
    // data structures for this view, not the model itself. This points up
    // the fact that, in the YACL implementation, even if the model is
    // shared by multiple views, the window system data structures
    // themselves are not shared between those views.
}


UI_WindowClass UI_TreeView::WindowClass () const
{
#if defined(__OS2__)
    return WC_CONTAINER;
#elif defined(__MS_WIN32__)
    return WC_TREEVIEW;
#else
    return 0;
#endif
}

void UI_TreeView::SetIndentDepth (short nPixels)
{
    _indent = nPixels;
    if (!ViewHandle())
        return;
#if defined(__OS2__)
    CNRINFO cnrInfo;
    cnrInfo.cb = sizeof (CNRINFO);
    cnrInfo.cxTreeIndent  = _indent;
    cnrInfo.cyLineSpacing = _separation;
    WinSendMsg (_handle, CM_SETCNRINFO, MPFROMP(&cnrInfo), MPFROMLONG
                (CMA_CXTREEINDENT));
#endif
}

void UI_TreeView::SetSeparation (short nPixels)
{
    _separation = nPixels;
    if (!ViewHandle())
        return;
#if defined(__OS2__)
    CNRINFO cnrInfo;
    cnrInfo.cb = sizeof (CNRINFO);
    cnrInfo.flWindowAttr  = CV_TREE | CV_ICON | CA_TREELINE;
    cnrInfo.cxTreeIndent  = _indent;
    cnrInfo.cyLineSpacing = _separation;
    WinSendMsg (_handle, CM_SETCNRINFO, MPFROMP(&cnrInfo), MPFROMLONG
                (CMA_LINESPACING));
#endif
}


bool UI_TreeView::IsExpanded (CL_TreeNodeLabel nodeId)
{
#if defined(__OS2__)
    if (!_handle)
        return FALSE;
    PRECORDCORE pRec = (PRECORDCORE) _recordCoreMap[nodeId];
    return pRec && (pRec->flRecordAttr & CRA_EXPANDED) ? TRUE : FALSE;
#endif
}


bool UI_TreeView::ExpandNode (CL_TreeNodeLabel nodeId)
{
#if defined(__OS2__)
    if (_handle) {
        PRECORDCORE pRec = (PRECORDCORE) _recordCoreMap[nodeId];
        if (pRec)
            return WinSendMsg (_handle, CM_EXPANDTREE,
                               MPFROMP (pRec), 0)
                ? TRUE : FALSE;
    }
#endif
    return FALSE;
}


bool UI_TreeView::CollapseNode (CL_TreeNodeLabel nodeId)
{
#if defined(__OS2__)
    if (_handle) {
        PRECORDCORE pRec = (PRECORDCORE) _recordCoreMap[nodeId];
        if (pRec)
            return  WinSendMsg (_handle, CM_COLLAPSETREE,
                                MPFROMP (pRec), 0)
                ? TRUE : FALSE;
    }
#endif
    return FALSE;
}







bool UI_TreeView::MakeVisualElement ()
{
    bool b = UI_SimpleVObject::MakeVisualElement ();

#if defined(__OS2__)
    CNRINFO cnrInfo;
    cnrInfo.cb = sizeof (CNRINFO);
//    cnrInfo.flWindowAttr = CV_TREE | CV_ICON | CA_TREELINE;
    cnrInfo.flWindowAttr = CV_TREE | CV_TEXT | CA_TREELINE;
    cnrInfo.cxTreeIndent = _indent;
    cnrInfo.cxTreeLine   = -1;
    cnrInfo.cyLineSpacing = _separation;
    WinSendMsg (_handle, CM_SETCNRINFO, MPFROMP(&cnrInfo), MPFROMLONG
                (CMA_FLWINDOWATTR | CMA_LINESPACING | CMA_CXTREEINDENT));
#endif
    
    // Now populate the tree.
    if (b && _model) {
        UI_TreeViewModel* tree = (UI_TreeViewModel*) _model;
        CL_ObjectSequence sq(1);
        CL_ObjectPtrTreeNode* root = tree->Root();
        if (root) {
            sq(0) = root;
            _AddChildren (sq, NULL);
        }
    }
    return b;
}







bool UI_TreeView::HandleEvent (UI_Event* e)
{
    if (e && e->Type() == Event_Select)
        _selectedNode = ((UI_TreeViewEvent*) e)->NodeId();
    return ProcessEvent(e);
}




#if defined(__OS2__)

// We need this StringDup function under OS/2, because we make a copy of
// the node text into the recordcore structure.
#include <string.h>
static char* _StringDup (const CL_String& s)
{
    char* dup = new char [s.Size() + 1];
    strcpy (dup, s.AsPtr());
    return dup;
}
#endif

    



void UI_TreeView::_AddChildren (const CL_ObjectSequence& children,
                                void* parent)
{
#if defined(__OS2__)
    TreeItem* firstItem = (TreeItem*) WinSendMsg
        (_handle, CM_ALLOCRECORD, MPFROMLONG (ExtraBytes),
         MPFROMSHORT ( (short) (children.Size())));
    TreeItem* p = firstItem;
    long n = children.Size();
    CL_GenericSequence dataSeq (n);
    long i;
    for (i = 0; i < n; i++) {
        dataSeq(i) = p;
        CL_ObjectPtrTreeNode* child = (CL_ObjectPtrTreeNode*) children(i);
        CL_ObjectPtr contentPtr = child->Content();
        if (contentPtr) {
            p->core.pszTree = _StringDup (contentPtr->AsString());
        }
        else {
            p->core.pszTree = NULL;
        }
        p->nodeId = child->Label();
        _recordCoreMap.Add (p->nodeId, (long) p);
        p = (TreeItem*) p->core.preccNextRecord;
    }
    RECORDINSERT ri;
    ri.cb                 = sizeof (RECORDINSERT);
    ri.pRecordOrder       = (PRECORDCORE) CMA_END;
    ri.pRecordParent      = (PRECORDCORE) parent;
    ri.zOrder             = (USHORT) CMA_TOP;
    ri.cRecordsInsert     = n;
    ri.fInvalidateRecord  = TRUE;
    WinSendMsg (_handle, CM_INSERTRECORD, MPFROMP(firstItem),
                MPFROMP(&ri));
    for (i = 0; i < n; i++) {
        CL_ObjectPtrTreeNode* child = (CL_ObjectPtrTreeNode*) children(i);
        if (child && !child->IsLeaf())
            _AddChildren (child->Children(), dataSeq(i));
    }
#endif
}










bool UI_TreeView::NodeExpanded (long /* nodeId */)
{
    return FALSE;
}


bool UI_TreeView::NodeCollapsed (long /* nodeId */)
{
    return FALSE;
}




#if defined(__OS2__)
long UI_TreeView::NodeIdFromMessage (QMSG& msg)
{
    if (msg.msg != WM_CONTROL)
        return -1;
    PRECORDCORE corePtr = NULL;
    USHORT notifyCode = SHORT2FROMMP (msg.mp1);
    if (notifyCode == CN_COLLAPSETREE || notifyCode == CN_EXPANDTREE)
        corePtr = (PRECORDCORE) msg.mp2;
    else if (notifyCode == CN_EMPHASIS) {
        PNOTIFYRECORDEMPHASIS ed = (PNOTIFYRECORDEMPHASIS) msg.mp2;
        if (ed)
            corePtr = ed->pRecord;
    }
    // Since we've asked for 4 extra bytes at allocation time, for the
    // node id, we can use them now:
    long nodeId = corePtr ? ((TreeItem*) corePtr)->nodeId : -1;
    return nodeId;
}

#endif


void UI_TreeView::ModelHasAddedNode (CL_ObjectPtrTreeNode* childNode)
{
#if defined(__OS2__)
    TreeItem* childRec = (TreeItem*) WinSendMsg
        (_handle, CM_ALLOCRECORD, MPFROMLONG (ExtraBytes), MPFROMLONG (1L));
    if (!childRec)
        return;
    if (childNode) {
        CL_ObjectPtr contentPtr = childNode->Content();
        CL_String labelText;
        if (contentPtr) {
            childRec->core.pszTree = _StringDup (contentPtr->AsString());
        }
        else
            childRec->core.pszTree = NULL;
        childRec->nodeId = childNode->Label();
        _recordCoreMap.Add (childNode->Label(), (long) childRec);
        CL_ObjectPtrTreeNode* parentNode = childNode->Parent();
        if (!parentNode) {
            CL_Error::Warning ("UI_TreeView::ModelHasAddedNode: "
                               "huh? no parent");
            return;
        }
        PRECORDCORE leftSibling = (PRECORDCORE) CMA_FIRST;
        long index = childNode->IndexInParent();
        if (index >= 1) {
            leftSibling = (PRECORDCORE) _recordCoreMap [
                parentNode->Child(index-1)->Label()];
        }
        CL_TreeNodeLabel parentLbl = parentNode->Label();
        RECORDINSERT ri;
        ri.cb                 = sizeof (RECORDINSERT);
        ri.pRecordOrder       = leftSibling;
        ri.pRecordParent      = (PRECORDCORE) _recordCoreMap [parentLbl];
        ri.zOrder             = (USHORT) CMA_TOP;
        ri.cRecordsInsert     = 1;
        ri.fInvalidateRecord  = TRUE;
        WinSendMsg (_handle, CM_INSERTRECORD, MPFROMP(childRec),
                    MPFROMP(&ri));
    }
#endif
}

void UI_TreeView::ModelHasNewRoot ()
{
    // Not yet implemented
}



void UI_TreeView::ModelWillDestroySubtree (CL_ObjectPtrTreeNode* node)
{
    _DoDestroy (node);
}


void UI_TreeView::_DoDestroy (CL_ObjectPtrTreeNode* node)
{
#if defined(__OS2__)
    if (!node)
        return;
    CL_ObjectPtrTreePostWalker walker (node);
    while (walker.More()) {
        CL_ObjectPtrTreeNode* nextNode = walker.Next();
        CL_TreeNodeLabel label = nextNode->Label();
        PRECORDCORE recCore = (PRECORDCORE) _recordCoreMap [label];
        if (recCore) {
            delete [] recCore->pszTree;
            WinSendMsg (_handle, CM_REMOVERECORD, &recCore,
                        MPFROM2SHORT(1, CMA_FREE | CMA_INVALIDATE));
        }
        _recordCoreMap.Remove (label);
    }
#endif
}




void UI_TreeView::ModelNodeContentChanged (CL_ObjectPtrTreeNode* node)
{
#if defined(__OS2__)
    if (!node)
        return;
    CL_ObjectPtr content = node->Content ();
    if (!content)
        return;
    PRECORDCORE recCore = (PRECORDCORE) _recordCoreMap [node->Label()];
    if (!recCore)
        return;
    if (recCore->pszTree)
        delete [] recCore->pszTree;
    recCore->pszTree = _StringDup (content->AsString());
    BOOL rc = (BOOL) WinSendMsg (_handle, CM_INVALIDATERECORD,
                                 MPFROMP(&recCore),
                                 MPFROM2SHORT (1, CMA_TEXTCHANGED));
    if (!rc) 
        UI_Application::PMError (__FILE__, __LINE__);
#endif
}

